/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Author: Norm Paxton (npaxton@novell.com)
******************************************************************************/
#include <cmpiutil/base.h>
#include <cmpiutil/string.h>
#include <stdio.h>
#include <errno.h>

#include <cmpiutil/cmpiUtils.h>

#include <cmpidt.h>
#include <cmpift.h>
#include <cmpimacs.h>

#define LOCAL_DEBUG

#ifdef LOCAL_DEBUG
#define DEBUGOUT(fmt, args...) printf(fmt,## args)
#else
#define DEBUGOUT(fmt, args...)
#endif


/*****************************************************************************/
CMPIObjectPath*
cmpiutilCreateCSObjectPath(
		const CMPIBroker * _broker,
		const char *ns,
		CMPIStatus* pstatus)
{
	CMPIObjectPath* cscop;
	if (!_broker || !ns || !pstatus)
	{
		if (pstatus)
		{
			CMSetStatusWithChars(_broker, pstatus, CMPI_RC_ERR_FAILED,
				"Invalid parms for cmpiutilCreateObjectPath");
		}
		errno = EINVAL;
		return NULL;
	}

	CMSetStatus(pstatus, CMPI_RC_OK);

	// Create and populate cs object path
	cscop = CMNewObjectPath(_broker,
							ns,
							(char *)cmpiutilGetComputerSystemClassName(),
							pstatus);
	if ((pstatus->rc != CMPI_RC_OK) || CMIsNullObject(cscop))
	{
		CMPIUTIL_SETSTATUS(_broker, pstatus, CMPI_RC_ERR_FAILED,
			"Could not create computer system cop");
	}
	else
	{
		// Add key to new object path
		char name[128];
		cmpiutilGetComputerSystemName(name, sizeof(name));

		CMAddKey(cscop,
				"CreationClassName",
				(CMPIValue *)(char *)cmpiutilGetComputerSystemClassName(),
				CMPI_chars);
		CMAddKey(cscop,
				"Name",
				(CMPIValue *)&name,
				CMPI_chars);
	}

	return cscop;
}

/*****************************************************************************/
CMPIObjectPath*
cmpiutilCreateOSObjectPath(
		const CMPIBroker* _broker,
		const char* ns,
		CMPIStatus* pstatus)
{
	CMPIObjectPath* objectpath;
	if (!_broker || !ns || !pstatus)
	{
		if (pstatus)
		{
			CMSetStatusWithChars(_broker, pstatus, CMPI_RC_ERR_FAILED,
				"Invalid parms for cmpiutilCreateObjectPath");
		}
		errno = EINVAL;
		return NULL;
	}

	CMSetStatus(pstatus, CMPI_RC_OK);

	// Create and populate OS object path
	objectpath = CMNewObjectPath(_broker, ns, (char *)cmpiutilGetOperatingSystemClassName(), pstatus);
	if ((pstatus->rc != CMPI_RC_OK) || CMIsNullObject(objectpath))
	{
		CMPIUTIL_SETSTATUS(_broker, pstatus, CMPI_RC_ERR_FAILED,
			"Could not create OS cop");
	}
	else
	{
		// Add key to new object path
		char hname[256];
		if (!cmpiutilGetComputerSystemName(hname, sizeof(hname)))
		{
			strncpy(hname, "localhost", sizeof(hname));
		}
		CMAddKey(objectpath, "CSCreationClassName", (CMPIValue*)cmpiutilGetComputerSystemClassName(), CMPI_chars);
		CMAddKey(objectpath, "CSName", (CMPIValue*)hname, CMPI_chars);
		CMAddKey(objectpath, "CreationClassName", (CMPIValue*)cmpiutilGetOperatingSystemClassName(), CMPI_chars);
		//CMAddKey(objectpath, "Name", (CMPIValue*)"Linux", CMPI_chars);
        // NWP: Changing to same value as CSName - as this is now using the Sblim-cmpi-base standards,
        // and we're creating a path to the sblim-cmpi-base/OperatingSystem
		CMAddKey(objectpath, "Name", (CMPIValue*)hname, CMPI_chars);
	}

	return objectpath;
}

/*****************************************************************************/
CMPIObjectPath*
cmpiutilCreateAssociationPath(
		const CMPIBroker * _broker,
		const char *ns,
		const char *className,
		const char *propName1,
		const char *propName2,
		const CMPIObjectPath *propRef1,
		const CMPIObjectPath *propRef2,
		CMPIStatus* pstatus)
{
	CMPIObjectPath* assocCop;

	if (!_broker
		|| !ns
		|| !className
		|| !propName1
		|| !propName2
		|| !propRef1
		|| !propRef2
		|| !pstatus)
	{
		if (pstatus)
		{
			CMSetStatusWithChars(_broker, pstatus, CMPI_RC_ERR_FAILED,
				"Invalid parms for cmpiutilCreateAssociationPath");
		}
		errno = EINVAL;
		return NULL;
	}

	CMSetStatus(pstatus, CMPI_RC_OK);

	// Now create and populate association instance
	assocCop = CMNewObjectPath(_broker, ns, className, pstatus);
	if ((pstatus->rc != CMPI_RC_OK) || CMIsNullObject(assocCop))
	{
		CMPIUTIL_SETSTATUS(_broker, pstatus, CMPI_RC_ERR_FAILED,
			"Could not create association ObjectPath");
	}
	else {
		CMAddKey(assocCop, propName1, (CMPIValue *)&propRef1, CMPI_ref);
		CMAddKey(assocCop, propName2, (CMPIValue *)&propRef2, CMPI_ref);
	}
	return assocCop;
}

/*****************************************************************************/
CMPIStatus
cmpiutilSetStatus(
	const CMPIBroker *_broker,
	CMPIStatus* pstatus,
	CMPIrc rc,
	const char* msg,
	const char* fileName,
	int lineno)
{
	char bfr[512];

	if (pstatus->rc == CMPI_RC_OK)
	{
		snprintf(bfr, sizeof(bfr)-1, "%s [file: %s line: %d]", msg, fileName, lineno);
	}
	else
	{
		const char *p = (pstatus->msg) ? CMGetCharPtr(pstatus->msg) : NULL;
		rc = pstatus->rc;
		if (p && *p)
		{
			snprintf(bfr, sizeof(bfr)-1, "%s : %s [file: %s line: %d]", p, msg, fileName, lineno);
		}
		else
		{
			snprintf(bfr, sizeof(bfr)-1, "%s [file: %s line: %d]", msg, fileName, lineno);
		}
	}
	bfr[sizeof(bfr)-1] = 0;
	CMSetStatusWithChars(_broker, pstatus, rc, bfr);
	return *pstatus;
}

/*****************************************************************************/
CMPIInstance*
cmpiutilCreateAssociationInst(
		const CMPIBroker * _broker,
		const char *ns,
		const char *className,
		const char **classKeys,
		const char ** properties,
		const char *propName1,
		const char *propName2,
		const CMPIObjectPath *propRef1,
		const CMPIObjectPath *propRef2,
		CMPIStatus* pstatus)
{
	CMPIInstance* assocInst;

	if (!_broker
		|| !ns
		|| !className
		|| !propName1
		|| !propName2
		|| !propRef1
		|| !propRef2
		|| !pstatus)
	{
		if (pstatus)
		{
			CMSetStatusWithChars(_broker, pstatus, CMPI_RC_ERR_FAILED,
				"Invalid parms for cmpiutilCreateAssociationInst");
		}
		errno = EINVAL;
		return NULL;
	}

	CMSetStatus(pstatus, CMPI_RC_OK);

	assocInst = cmpiutilNewInstance(_broker, ns, className, classKeys, properties, pstatus);
	if ((pstatus->rc != CMPI_RC_OK) || CMIsNullObject(assocInst))
	{
		CMPIUTIL_SETSTATUS(_broker, pstatus, CMPI_RC_ERR_FAILED,
			"Could not create association instance");
	}
	else
	{
		CMSetProperty(assocInst, propName1, (CMPIValue *)&propRef1, CMPI_ref);
		CMSetProperty(assocInst, propName2, (CMPIValue *)&propRef2, CMPI_ref);
	}
	return assocInst;
}

/*****************************************************************************/
CMPIBoolean
cmpiutilClassIsDerivedFrom(
		const char *derivedClass,
		const char *superClass,
		const CMPIBroker * _broker,
		const char *ns,
		CMPIStatus* pstatus)
{
	CMPIBoolean bRetVal = 0;

	if (!derivedClass
		|| !superClass
		|| !_broker
		|| !ns
		|| !pstatus)
	{
		if (pstatus)
		{
			CMSetStatusWithChars(_broker, pstatus, CMPI_RC_ERR_FAILED,
				"Invalid parms for cmpiutilClassIsDerivedFrom");
		}
		errno = EINVAL;
		return 0;
	}

	CMSetStatus(pstatus, CMPI_RC_OK);
	if (strcasecmp(derivedClass, superClass) == 0)
	{
		return 1;
	}

	CMPIObjectPath* childClass = CMNewObjectPath(_broker,
												ns,
												derivedClass,
												pstatus);
	if ((pstatus->rc != CMPI_RC_OK) || CMIsNullObject(childClass))
	{
		CMPIUTIL_SETSTATUS(_broker, pstatus, CMPI_RC_ERR_FAILED,
			"Could not create ObjectPath");
		return 0;
	}

	bRetVal = CMClassPathIsA(_broker, childClass, superClass, pstatus);
	return bRetVal;
}

/*****************************************************************************/
CMPIInstance*
cmpiutilNewInstance(
		const CMPIBroker *_broker,
		const char *ns,
		const char *className,
		const char** classKeys,
		const char** properties,
		CMPIStatus* pstatus)
{
	CMPIObjectPath *objectpath;
	CMPIInstance* instance;

	if (!_broker
		|| !ns
		|| !className
		|| !classKeys
		|| !pstatus)
	{
		if (pstatus)
		{
			CMSetStatusWithChars(_broker, pstatus, CMPI_RC_ERR_FAILED,
				"Invalid parms for cmpiutilmpiNewInstance");
		}
		errno = EINVAL;
		return NULL;
	}

	CMSetStatus(pstatus, CMPI_RC_OK);

	objectpath = CMNewObjectPath(_broker, ns, className, pstatus);
	if ((pstatus->rc != CMPI_RC_OK) || CMIsNullObject(objectpath))
	{
		CMPIUTIL_SETSTATUS(_broker, pstatus, CMPI_RC_ERR_FAILED,
			"Failed creating new object path");
		return NULL;
	}

	/* Create a new CMPIInstance */
	instance = CMNewInstance(_broker, objectpath, pstatus);
	if ((pstatus->rc != CMPI_RC_OK) || CMIsNullObject(instance))
	{
		CMPIUTIL_SETSTATUS(_broker, pstatus, CMPI_RC_ERR_FAILED,
			"Failed creating new instance");
		return NULL;
	}

	/* Setup a filter to only return the desired properties */
	*pstatus = CMSetPropertyFilter(instance, properties, classKeys);
	if (pstatus->rc != CMPI_RC_OK)
	{
		CMPIUTIL_SETSTATUS(_broker, pstatus, CMPI_RC_ERR_FAILED,
			"Cannot set property filter");
		return NULL;
	}

	return instance;
}

/*****************************************************************************/
CMPIBoolean
cmpiutilAreValuesSame(
		const CMPIData val1,
		const CMPIData val2,
		CMPIStatus *pStatus)
{
	CMPIBoolean bRetVal = 0;

	if (val1.type == val2.type)  // if not same, return default 0
	{
		if ( (CMIsNullValue(val1) && !CMIsNullValue(val2)) ||
			 (!CMIsNullValue(val1) && CMIsNullValue(val2)) )
		{
			return 0;	// FALSE
		}

		switch (val1.type)
		{
			case CMPI_boolean:
				bRetVal = val1.value.boolean == val2.value.boolean;
				break;
			case CMPI_char16:
				bRetVal = val1.value.char16 == val2.value.char16;
				break;
			case CMPI_real32:
				bRetVal = val1.value.real32 == val2.value.real32;
				break;
			case CMPI_real64:
				bRetVal = val1.value.real64 == val2.value.real64;
				break;
			case CMPI_uint8:
				bRetVal = val1.value.uint8 == val2.value.uint8;
				break;
			case CMPI_uint16:
				bRetVal = val1.value.uint16 == val2.value.uint16;
				break;
			case CMPI_uint32:
				bRetVal = val1.value.uint32 == val2.value.uint32;
				break;
			case CMPI_uint64:
				bRetVal = val1.value.uint64 == val2.value.uint64;
				break;
			case CMPI_sint8:
				bRetVal = val1.value.sint8 == val2.value.sint8;
				break;
			case CMPI_sint16:
				bRetVal = val1.value.sint16 == val2.value.sint16;
				break;
			case CMPI_sint32:
				bRetVal = val1.value.sint32 == val2.value.sint32;
				break;
			case CMPI_sint64:
				bRetVal = val1.value.sint64 == val2.value.sint64;
				break;
			case CMPI_ref:
				bRetVal = cmpiutilAreObjectPathsSame(val1.value.ref,
													val2.value.ref,
													pStatus );
				break;
			case CMPI_string:
				bRetVal = (!strcmp(CMGetCharPtr(val1.value.string),
					               CMGetCharPtr(val2.value.string)));
				break;
			case CMPI_chars:
				bRetVal = (!strcmp(val1.value.chars, val2.value.chars));
				break;
			case CMPI_dateTime:
				bRetVal = ( CMGetBinaryFormat(val1.value.dateTime, pStatus) ==
					        CMGetBinaryFormat(val2.value.dateTime, pStatus) );
				break;
			//case CMPI_ptr:
			// don't know how to handle
			//case CMPI_charsptr:
			// don't know how to handle
			//
			// TODO: Add handling for arrays
			default:
				bRetVal = 0;
				break;
		}
	}
	return bRetVal;
}

/*****************************************************************************/
CMPIBoolean
cmpiutilAreObjectPathsSame(
		const CMPIObjectPath *cop1,
		const CMPIObjectPath *cop2,
		CMPIStatus *pStatus)
{
	DEBUGOUT("cmpiutilAreObjectPathsSame() called\n");

	if (!cop1 || !cop2)
	{
		DEBUGOUT("return FALSE at 1\n");
		errno = EINVAL;
		return 0;	// FALSE
	}

	
	DEBUGOUT("Comparing Object Paths:\n");
	DEBUGOUT("   cop1: %s\n", CMGetCharPtr(CMObjectPathToString(cop1, NULL)));
	DEBUGOUT("   cop2: %s\n", CMGetCharPtr(CMObjectPathToString(cop2, NULL)));
	

	// sanity checks - same namespace, class, and property list size
	char *hn1 = (char*) CMGetCharPtr(CMGetHostname(cop1, pStatus));
	char *hn2 = (char*) CMGetCharPtr(CMGetHostname(cop2, pStatus));
    if (hn1 != NULL)
    	DEBUGOUT("   hn1: %s\n", hn1);
    else
    	DEBUGOUT("   hn1 is NULL\n");
    if (hn2 != NULL)
	    DEBUGOUT("   hn2: %s\n", hn2);
    else
    	DEBUGOUT("   hn2 is NULL\n");

	char *ns1 = (char*) CMGetCharPtr(CMGetNameSpace(cop1, pStatus));
	char *ns2 = (char*) CMGetCharPtr(CMGetNameSpace(cop2, pStatus));
    if (ns1 != NULL)
    	DEBUGOUT("   ns1: %s\n", ns1);
    else
    	DEBUGOUT("   hn1 is NULL\n");
    if (ns2 != NULL)
    	DEBUGOUT("   ns2: %s\n", ns2);
    else
    	DEBUGOUT("   ns2 is NULL\n");

	char *cl1 = (char*) CMGetCharPtr(CMGetClassName(cop1, pStatus));
	char *cl2 = (char*) CMGetCharPtr(CMGetClassName(cop2, pStatus));

	unsigned int numkeys1 = CMGetKeyCount(cop1, pStatus);
	unsigned int numkeys2 = CMGetKeyCount(cop2, pStatus);

	if (hn1 && hn2 && *hn1 && *hn2)
	{
		// only compare if I have a hostname for each
		if (strcasecmp(hn1, hn2) != 0) // not same
		{
			DEBUGOUT("HostNames are different:  %s  !=  %s\n", hn1, hn2);
			return 0;	// FALSE
		}
	}
	if (ns1 && ns2 && *ns1 && *ns2)
	{
		// only compare if I have a namespace for each
		if (strcasecmp(ns1, ns2) != 0) // not same
		{
			DEBUGOUT("Namespaces are different:  %s  !=  %s\n", ns1, ns2);
			return 0;	// FALSE
		}
	}
    if ((cl1 && cl2 && *cl1 && *cl2) && (strcasecmp(cl1, cl2) != 0)) // not same
	{
		DEBUGOUT("ClassNames are different:  %s  !=  %s\n", cl1, cl2);
		return 0;	// FALSE
	}
	if (numkeys1 != numkeys2) // different number of keys
	{
		DEBUGOUT("Number of keys are different:  %d != %d\n",numkeys1,numkeys2);
		return 0;	// FALSE
	}

	// now go through each key, string-ize it, and compare
	// get keyname from first op, then use that keyname to get the key
	// from second op... they may not be in the same order
	CMPIData key1, key2;
	CMPIString *keyname;
	unsigned int i = 0;

	DEBUGOUT("Looping through %d keys in compare\n", numkeys1);
	for (i=0; i<numkeys1; i++)
	{
		key1 = CMGetKeyAt(cop1, i, &keyname, pStatus);
		key2 = CMGetKey(cop2, CMGetCharPtr(keyname), pStatus);

		DEBUGOUT("  key[%d]: keyname: %s,\n", i, CMGetCharPtr(keyname));

		if ( (CMIsNullValue(key1) && !CMIsNullValue(key2)) ||
			 (!CMIsNullValue(key1) && CMIsNullValue(key2)) )
		{
			return 0;	// FALSE
		}

		if (!cmpiutilAreValuesSame(key1, key2, pStatus))
		{
			return 0;	// FALSE
		}
	}

	return 1;	// TRUE
}




/**
 * Given a CMPIObjectPath, create a string-ified version of the
 * object path
 */
/*
CMPIString *
cmpiutilGetStringFromObjectPath(
		CMPIBroker *broker,
		const CMPIObjectPath *cop,
		CMPIStatus *pStatus )
{
	if ((cop == NULL) || CMIsNullObject(cop))
	{
		return NULL;
	}

	CMPIString *str = NULL;
	char stringCop[4096] = {0};

	CMPIString *hostname = CMGetHostname(cop, pStatus);
	CMPIString *ns = CMGetNameSpace(cop, pStatus);
	CMPIString *classname = CMGetClassName(cop, pStatus);
	unsigned int keyCount = CMGetKeyCount(cop, pStatus);
	unsigned int i=0;
	char host[128] = {0};

	if (!CMIsNullObject(hostname))
	{
		strncpy(host, CMGetCharPtr(hostname), 128);
	}
	else
	{
		cmpiutilGetComputerSystemName(host, 128);
	}

	if (!*host)
		return NULL;

	if (CMIsNullObject(ns))
		return NULL;

	if (CMIsNullObject(classname))
		return NULL;

	sprintf(stringCop, "//%s/%s:%s.", host, CMGetCharPtr(ns), CMGetCharPtr(classname));

	for (i=0; i<keyCount; i++)
	{
		CMPIString *name;
		CMPIData keydata = CMGetKeyAt(cop, i, &name, pStatus);
		CMPIValue keyval = keydata.value;

		char strVal[1024] = {0};

		switch (keydata.type)
		{
			case CMPI_boolean:
				sprintf(strVal, "%d", (int)keyval.boolean);
				break;
			case CMPI_char16:
				sprintf(strVal, "%d", (int)keyval.char16);
				break;
			case CMPI_real32:
				sprintf(strVal, "%f", (int)keyval.real32);
				break;
			case CMPI_real64:
				sprintf(strVal, "%f", (int)keyval.real64);
				break;
			case CMPI_uint8:
				sprintf(strVal, "%d", (int)keyval.uint8);
				break;
			case CMPI_uint16:
				sprintf(strVal, "%d", (int)keyval.uint16);
				break;
			case CMPI_uint32:
				sprintf(strVal, "%d", (int)keyval.uint32);
				break;
			case CMPI_uint64:
				sprintf(strVal, "%u", (int)keyval.uint64);
				break;
			case CMPI_sint8:
				sprintf(strVal, "%d", (int)keyval.sint8);
				break;
			case CMPI_sint16:
				sprintf(strVal, "%d", (int)keyval.sint16);
				break;
			case CMPI_sint32:
				sprintf(strVal, "%d", (int)keyval.sint32);
				break;
			case CMPI_sint64:
				sprintf(strVal, "%d", (int)keyval.sing64);
				break;
			case CMPI_ref:
				sprintf(strVal, "%s", CMGetCharPtr(
							cmpiutilGetStringFromObjectPath(keyval.ref,
													pStatus )));
				break;
			case CMPI_string:
				sprintf(strVal, "%s", (char *)CMGetCharPtr(keyval.string));
				break;
			case CMPI_chars:
				sprintf(strVal, "%s", (char *)(keyval.chars));
				break;
			case CMPI_dateTime:
				sprintf(strVal, "%s", (char *)CMGetStringFormat(keyval.dateTime, pStatus));
				break;
			//case CMPI_ptr:
			// don't know how to handle
			//case CMPI_charsptr:
			// don't know how to handle
			//
			// TODO: Add handling for arrays
			default:
				bRetVal = 0;
				break;
		}


		char tmpStrVal[4096] = {0};
		snprintf(tmpStrVal, 4096, "%s=\\\"%s\\\"", CMGetCharPtr(name), strVal);
		strncat(stringCop, tmpStrVal, 4096-strlen(strVal));

        str = CMNewString(broker, stringCop, pStatus);
	}
	return str;
}
*/

