/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Author: Jon Carey (jcarey@novell.com)
******************************************************************************/
#include <cmpiutil/exec.h>
#include <cmpiutil/base.h>
#include <cmpiutil/string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <errno.h>
#include <poll.h>
#include <time.h>

#ifndef NSIG
#define NSIG 64
#endif

#ifndef MIN
#undef MIN
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

static pid_t safeWaitPid(pid_t pid, int* status, int options);
static pid_t noIntrWaitPid(pid_t pid, int* status, int options);
static int timedWaitPid(pid_t pid, int* pstatus, uint32_t waitTime);
static int killWait(pid_t pid, int* pstatus, uint32_t waitTime, int sig);
static int terminateChild(pid_t pid, int* pstatus);
static int waitForInput2(int fd1, int fd2, int timeoutSecs);
static int waitForInput(int fd, int timeoutSecs);
static int localWrite(int fd, const void* buf, size_t count);
static int writeStr(int fd, const char* str);
static int localRead(int fd, void* buf, size_t count);
static int closePipe(int fds[2], int cc);
static int openPipe(int fds[2]);
static void writeErrno(int fd);
static int appendToBuffer(char** outBfr, size_t *outBfrSize,
	size_t outOffset, int outBfrSizeLimit, char* data,
	unsigned int dataLen);

/*****************************************************************************/
int
cmpiutilSafeSystem(
	char* const argv[],
	char* const envp[])
{
	int status;
	pid_t pid;

	if (!argv) 
	{
		errno = EINVAL;
		return -1;
	}

	pid = fork();
	if (pid == -1) 
	{
		return -1;
	}

	if (pid == 0) 	/* Child */
	{
		struct rlimit rl;
		int i;
		sigset_t emptymask;
		size_t sig;

		sigemptyset(&emptymask);
		sigprocmask(SIG_SETMASK, &emptymask, 0);
		for(sig = 1; sig <= NSIG; ++sig) 
		{
			struct sigaction temp;
			sigaction(sig, 0, &temp);
			temp.sa_handler = SIG_DFL;
			sigaction(sig, &temp, NULL);
		}

		i = sysconf(_SC_OPEN_MAX);
		if (getrlimit(RLIMIT_NOFILE, &rl) != -1) 
		{
			i = (i < 0) ? rl.rlim_max : MIN(rl.rlim_max, i);
		}

		/* Set close on exec for all child file handles */
		while (i > 2) 
		{
			fcntl(i, F_SETFD, FD_CLOEXEC);
			i--;
		}

		if (envp) 
		{
			execve(argv[0], argv, envp);
		}
		else 
		{
			execv(argv[0], argv);
		}
		perror("Failure on exec:");
		_exit(127);
	}

	while (1) 
	{
		if (waitpid(pid, &status, 0) == -1) 
		{
			if (errno != EINTR) 
			{
				return -1;
			}
		}
		else 
		{
			status = WEXITSTATUS(status);
			break;
		}
	}

	return status;
}

/*****************************************************************************/
int
cmpiutilSafePopen(
	char* const argv[],
	char* const envp[],
	int* childsStdin,
	int* childsStdout,
	int* childsStderr)
{
	int stdinfds[2], stdoutfds[2], stderrfds[2], execerrfds[2];
	int childErrorCode, cc;

	if (!argv || !childsStdin || !childsStdout || !childsStderr) 
	{
		errno = EINVAL;
		return -1;
	}

	if (openPipe(stdinfds) == -1) 
	{
		return -1;
	}

	if (openPipe(stdoutfds) == -1) 
	{
		return closePipe(stdinfds, -1);
	}

	if (openPipe(stderrfds) == -1) 
	{
		closePipe(stdinfds, 0);
		return closePipe(stdoutfds, -1);
	}

	if (openPipe(execerrfds) == -1) 
	{
		closePipe(stderrfds, 0);
		closePipe(stdinfds, 0);
		return closePipe(stdoutfds, -1);
	}

	pid_t forkrv = fork();
	if (forkrv == -1) 
	{
		closePipe(stdinfds, 0);
		closePipe(stdoutfds, 0);
		closePipe(stderrfds, 0);
		return closePipe(execerrfds, -1);
	}

	if (forkrv == 0) /* child */
	{
		struct rlimit rl;
		int i;
		sigset_t emptymask;
		size_t sig;

		sigemptyset(&emptymask);
		sigprocmask(SIG_SETMASK, &emptymask, 0);
		for(sig = 1; sig <= NSIG; ++sig) 
		{
			struct sigaction temp;
			sigaction(sig, 0, &temp);
			temp.sa_handler = SIG_DFL;
			sigaction(sig, &temp, NULL);
		}
		close(0);
		close(1);
		close(2);
		close(execerrfds[0]);	/* Won't need this */

		/* Setup stdin and close stdin pipe handles */
		if (dup2(stdinfds[0], 0) == -1) 
		{
			writeErrno(execerrfds[1]);
			_exit(127);
		}
		closePipe(stdinfds, 0);

		/* Setup stdout and close stdout pipe handles */
		if (dup2(stdoutfds[1], 1) == -1) 
		{
			writeErrno(execerrfds[1]);
			_exit(127);
		}
		closePipe(stdoutfds, 0);

		/* Setup stderr and close stderr pipe handles */
		if (dup2(stderrfds[1], 2) == -1) 
		{
			writeErrno(execerrfds[1]);
			_exit(127);
		}
		closePipe(stderrfds, 0);

		i = sysconf(_SC_OPEN_MAX);
		if (getrlimit(RLIMIT_NOFILE, &rl) != -1) 
		{
			i = (i < 0) ? rl.rlim_max : MIN(rl.rlim_max, i);
		}

		/* Set close on exec for all child file handles */
		while (i > 2) 
		{
			fcntl(i, F_SETFD, FD_CLOEXEC);
			i--;
		}

		if (envp) 
		{
			execve(argv[0], argv, envp);
		}
		else 
		{
			execv(argv[0], argv);
		}
		/* If we made it here then exec.. failed, so we'll write errno to
		 * the error pipe.
		 */
		writeErrno(execerrfds[1]);
		close(execerrfds[1]);
		_exit(127);
	}

	/* Prevent the parent from using the child's end of the pipes */
	close(stdinfds[0]);
	close(stdoutfds[1]);
	close(stderrfds[1]);

	/* close the parent's output side so that when the clild's output side is
	 * closed, it can be detected.
	 */
	close(execerrfds[1]);

	cc = waitForInput(execerrfds[0], 10);
	if (cc == 0) /* timed out? */
	{
		/* For some reason the child never ran exec(). Must've deadlocked or
		 * the system is *really* buzy. Brutally kill it.
		 */
		kill(-forkrv, SIGKILL); // send SIGKILL to all processes in the process group
		errno = ETIMEDOUT;
		cc = -1;
	}
	else 
	{
		cc = 0;
		if (localRead(execerrfds[0], &childErrorCode,
			sizeof(childErrorCode)) > 0) 
		{
			errno = childErrorCode;
			cc = -1;
		}
	}
	close(execerrfds[0]);

	if (cc < 0) 
	{
		int state;
		close(stdinfds[1]);
		close(stdoutfds[0]);
		close(stderrfds[0]);
		*childsStdin = *childsStdout = *childsStderr = -1;
		noIntrWaitPid(forkrv, &state, 0); /* Clean up child proc */
	}
	else 
	{
		cc = forkrv;	/* Set return value to pid of child process */
		*childsStdin = stdinfds[1];
		*childsStdout = stdoutfds[0];
		*childsStderr = stderrfds[0];
	}

	return cc;
}

/*****************************************************************************/
int
cmpiutilExecuteProcessAndGatherOutputAndError(
	char* const argv[],
	char* const envp[],
	char** output,
	size_t* outputSize,
	int outputLimit,
	char** errOutput,
	size_t* errOutputSize,
	int errOutputLimit,
	const char* input,
	int* processStatus,
	int timeOutSecs)
{
	int cpid, cc, rerrno;
	int childsStdin, childsStdout, childsStderr;
	char *outBfr, *errBfr, readBfr[512];
	size_t outBfrWrote, errBfrWrote, outAllocSize, errAllocSize;
	size_t bytesRead;
	uint8_t errOpened, outOpened;
	uint8_t isStdErr = (!errOutput) ? 0 : 1;

	/* Validate the parameters */
	if (!argv || !output || !outputSize || !processStatus
		|| (isStdErr && !errOutputSize)) 
	{
		errno = EINVAL;
		return -1;
	}

	outBfr = errBfr = 0;
	*outputSize = outBfrWrote = errBfrWrote = 0;
	*output = 0;

	if (isStdErr) 
	{
		*errOutput = 0;
		*errOutputSize = 0;
	}

	rerrno = cc = 0;
	for(;;) 
	{
		cpid = cmpiutilSafePopen(argv, envp, &childsStdin, &childsStdout,
			&childsStderr);
		if (cpid < 0) 
		{
			rerrno = errno;
			cc = -1;
			break;
		}

		if (input) 
		{
			/* Write string to childs stdin */
			if (writeStr(childsStdin, input) == -1) 
			{
				rerrno = errno;
				cc = -1;
				break;
			}
		}

		errOpened = outOpened = 1;
		while (errOpened || outOpened) 
		{
			int wcc = waitForInput2(childsStdout, childsStderr, timeOutSecs);
			if (wcc <= 0) 
			{
				/* Well the child proc has exceeded its allowed time to
				 * respond or we failed waiting for a response, so attempt
				 * to kill it.
				 */
				rerrno = (wcc == 0) ? ETIMEDOUT : errno;
				terminateChild(cpid, processStatus);
				//cpid = cc = -1;
				cpid = -1;
				break;
			}

			/* Read the child's stdout */
			if (wcc == 1 || wcc == 3) 
			{
				bytesRead = localRead(childsStdout, readBfr, sizeof(readBfr));
				if (bytesRead > 0) 
				{
					if (!outBfr) 
					{
						/* Initial memory allocation */
						outAllocSize = (outputLimit > 0) ? MIN(outputLimit,1024) : 1024;
						outBfr = (char*)malloc(outAllocSize);
						if (!outBfr) 
						{
							if (errBfr)
								free(errBfr);
							return -1;
						}
					}
					if (appendToBuffer(&outBfr, &outAllocSize, outBfrWrote,
						outputLimit, readBfr, bytesRead) == -1) 
					{
						/* Memory allocation error or we have exceed the
						 * output limit
						 */
						rerrno = errno;
						cc = -1;
						break;
					}
					outBfrWrote += bytesRead;
				}
				else 
				{
					outOpened = 0;
				}
			}

			/* Read the child's stderr */
			if (wcc > 1) 
			{
				bytesRead = localRead(childsStderr, readBfr, sizeof(readBfr));
				if (bytesRead > 0) 
				{
					/* Allocate the initial std bfr if it hasn't been done */
					if (isStdErr) 
					{
						if (!errBfr) 
						{
							errAllocSize = (errOutputLimit > 0) ? MIN(errOutputLimit,1024) : 1024;
							errBfr = (char*)malloc(errAllocSize);
							if (!errBfr) 
							{
								if (outBfr)
									free(outBfr);
								errno = ENOMEM;
								return -1;
							}
						}

						if (appendToBuffer(&errBfr, &errAllocSize, errBfrWrote,
							errOutputLimit, readBfr, bytesRead) == -1) 
						{
							/* Memory allocation error or we have exceed the
							 * output limit
							 */
							rerrno = errno;
							cc = -1;
							break;
						}
						errBfrWrote += bytesRead;
					}
				}
				else 
				{
					errOpened = 0;
				}
			}
		}

		close(childsStdout);
		close(childsStdin);
		close(childsStderr);

		/* Clean up child proc and get status */
		if (cpid > 0) 
		{
			if (timedWaitPid(cpid, processStatus, timeOutSecs) != 0) 
			{
				/* Exceeded time getting status */
				rerrno = ETIMEDOUT;
				terminateChild(cpid, processStatus);
				cc = -1;
			}
		}

		break;
	}

	if (cc < 0) 
	{
		free(outBfr);
		if (isStdErr) 
		{
			free(errBfr);
		}
	}
	else 
	{
		/* On success, the caller is responsible for deallocating any memory
		 * we have allocated during this operation
		 */
		*outputSize = outBfrWrote;
		*output = outBfr;
		if (isStdErr) 
		{
			*errOutputSize = errBfrWrote;
			*errOutput = errBfr;
		}
	}
	errno = rerrno;
	return cc;
}

/*****************************************************************************/
static pid_t
safeWaitPid(pid_t pid, int* status, int options)
{
	int localReturnValue = -1;
	pid_t returnPID = waitpid(pid, &localReturnValue, options);
	if (returnPID > 0) 
	{
		*status = localReturnValue;
	}
	return returnPID;
}

/*****************************************************************************/
static pid_t
noIntrWaitPid(pid_t pid, int* status, int options)
{
	pid_t waitpidrv;
	do 
	{
		waitpidrv = safeWaitPid(pid, status, options);
	} while (waitpidrv == -1 && errno == EINTR);
	return waitpidrv;
}

/*****************************************************************************/
static int
timedWaitPid(
	pid_t pid,
	int* pstatus,
	uint32_t waitTime)	/* Time in seconds */
{
	unsigned int N = 154;
	unsigned int M = 128;  /* N/M is about 1.20 */
	unsigned int MAXPERIOD = 5000;
	unsigned int period = 100;
	unsigned int t = 0;
	/* conver given time to ms */
	waitTime *= 1000;
	pid_t waitpidrv = noIntrWaitPid(pid, pstatus, WNOHANG);

	/* waitpiddrv will be 0 if the child is still running.
	 * (i.e. status is not yet available)
	 */
	while (t < waitTime && waitpidrv == 0) 
	{
		cmpiutilMilliSleep(period);
		t += period;
		period *= N;
		period /= M;
		if (period > MAXPERIOD) 
		{
			period = MAXPERIOD;
		}
		waitpidrv = noIntrWaitPid(pid, pstatus, WNOHANG);
	}

	if (waitpidrv < 0) 
	{
		/* waitpid failed */
		return -1;
	}

	/* If waitpidrv is 0 at this point, we know the we have exceeded
	 * the waitTime parm and the process is still running.
	 */
	if (waitpid == 0) 
	{
		errno = ETIMEDOUT;
		return -1;
	}
	return 0;
}

/*****************************************************************************/
static int
killWait(
	pid_t pid,
	int* pstatus,
	uint32_t waitTime,
	int sig)
{
	if (kill(-pid, sig) == -1) // send sig to all processes in the process group (pid of the process group leader) 
	{
		/* don't trust waitpid, Format ctor, etc. to leave errno alone */
		int errnum = errno;
		/* maybe kill() failed because child terminated first */
		if (noIntrWaitPid(pid, pstatus, WNOHANG) > 0) 
		{
			return 0;
		}
		else 
		{
			errno = errnum;
			return -1;
		}
	}
	return timedWaitPid(pid, pstatus, waitTime);
}

/*****************************************************************************/
static int
terminateChild(
	pid_t pid,
	int* pstatus)
{
	if (killWait(pid, pstatus, 10, SIGTERM) != 0) 
	{
		return killWait(pid, pstatus, 10, SIGKILL);
	}
	return 0;
}

/*****************************************************************************
 * @return
 * 		-1 = Error occurred
 * 		0 = time out
 * 		1 = fd1 has input available
 * 		2 = fd2 has input available
 * 		3 = fd1 & fd2 have input available
 */
static int
waitForInput2(
	int fd1,
	int fd2,
	int timeoutSecs)
{
	int cc;
	struct pollfd pfds[2];
	int toms = (timeoutSecs < 0) ? -1 :	(timeoutSecs * 1000);

	do 
	{
		memset(pfds, 0, sizeof(pfds));
		pfds[0].fd = fd1;
		pfds[0].events = POLLIN;
		pfds[1].fd = fd2;
		pfds[1].events = POLLIN;
		cc = poll(pfds, 2, toms);
	} while (cc == -1 && (errno == EAGAIN || errno == EINTR));

	if (cc > 0) 
	{
		cc = 0;
		if (pfds[0].revents & (POLLIN | POLLHUP))
			cc = 1;
		if (pfds[1].revents & (POLLIN | POLLHUP))
			cc += 2;
	}

	return cc;
}

/*****************************************************************************/
static int
waitForInput(
	int fd,
	int timeoutSecs)
{
	int cc;
	struct pollfd pfd;
	int toms = (timeoutSecs < 0) ? -1 :	(timeoutSecs * 1000);

	do 
	{
		memset(&pfd, 0, sizeof(pfd));
		pfd.fd = fd;
		pfd.events |= POLLIN;
		cc = poll(&pfd, 1, toms);
	} while (cc == -1 && (errno == EAGAIN || errno == EINTR));

	return cc;
}

/*****************************************************************************/
static int
localWrite(
	int fd,
	const void* buf,
	size_t count)
{
	int byteswritten;
	do 
	{
		byteswritten = write(fd, buf, count);
	} while (byteswritten == -1 && errno == EINTR);
	return byteswritten;
}

/*****************************************************************************/
static int
writeStr(
	int fd,
	const char* str)
{
	return localWrite(fd, str, strlen(str));
}

/*****************************************************************************/
static int
localRead(
	int fd,
	void* buf,
	size_t count)
{
	int bytesread;
	do 
	{
		bytesread = read(fd, buf, count);
	} while(bytesread == -1 && errno == EINTR);
	return bytesread;
}

/*****************************************************************************/
static int
closePipe(
	int fds[2],
	int cc)
{
	int lerrno = errno;
	close(fds[0]);
	close(fds[1]);
	errno = lerrno;
	return cc;
}

/*****************************************************************************/
static int
openPipe(
	int fds[2])
{
	int i, fdflags;
	if (pipe(fds) == -1) 
	{
		return -1;
	}

	for (i = 0; i < 2; i++) 
	{
		fdflags = fcntl(fds[i], F_GETFL, 0);
		if (fdflags == -1) 
		{
			return closePipe(fds, -1);
		}
		fdflags &= O_NONBLOCK;
		if (fcntl(fds[i], F_SETFL, fdflags) == -1) 
		{
			return closePipe(fds, -1);
		}
	}
	return 0;
}

/*****************************************************************************/
static void
writeErrno(
	int fd)
{
	int lerrno = errno;
	localWrite(fd, &lerrno, sizeof(lerrno));
	errno = lerrno;
}

/*****************************************************************************/
static int
appendToBuffer(
	char** outBfr,
	size_t *outBfrSize,
	size_t outOffset,
	int outBfrSizeLimit,
	char* data,
	unsigned int dataLen)
{
	/* Ensure we're not going to exceed the limit */
	if (outBfrSizeLimit > 0) 
	{
		if ((outOffset + dataLen) > (outBfrSizeLimit -1)) 
		{
			/* Will exceed the buffer limit */
			errno = EOVERFLOW;
			return -1;
		}
	}

	/* Ensure we have enough memory */
	if ((outOffset + dataLen) > (*outBfrSize-1)) 
	{
		char* lbfr = realloc(*outBfr, (*outBfrSize)+1024);
		if (!lbfr) 
		{
			return -1;
		}
		*outBfr = lbfr;
		*outBfrSize += 1024;
	}
	memcpy((*outBfr)+outOffset, data, dataLen);
	*((*outBfr)+outOffset+dataLen) = 0;
	return 0;
}

