/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Author: Jon Carey (jcarey@novell.com)
******************************************************************************/
#include <cmpiutil/modifyFile.h>
#include <cmpiutil/string.h>
#include <cmpiutil/ioUtils.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <limits.h>

/****************************************************************************/
static int
errCond(
		char* errMsgBfr,
		size_t errMsgBfrLen,
		int cc,
		const char* fmt,
		...)
{
	va_list ap;

	if (!errMsgBfr || !errMsgBfrLen) 
	{
		return cc;
	}

	va_start(ap, fmt);
	vsnprintf(errMsgBfr, errMsgBfrLen, fmt, ap);
	va_end(ap);
	return cc;
}

/****************************************************************************/
int
cmpiutilModifyFile(
		const char* fileName,
		const cmpiutilFileModifier_CBS* const cbs,
		char* errMsgBfr,
		size_t errMsgBfrLen)
{
	int modified, cc, aborted = 0;
	size_t bfrLen = 0;
	char *p, *bufPtr = NULL;
	char tmpFileName[PATH_MAX], lerrMsgBfr[80];
	char dirName[PATH_MAX];
	FILE *tempHdl, *origHdl;
	struct stat statbuf;

	if (!fileName || !cbs) 
	{
		return errCond(errMsgBfr, errMsgBfrLen, -1, "Invalid parameters");
	}

	modified = 0;
	if (!errMsgBfr || !errMsgBfrLen) 
	{
		errMsgBfr = lerrMsgBfr;
		errMsgBfrLen = sizeof(lerrMsgBfr);
	}
	*errMsgBfr = '\0';

	if (access(fileName, W_OK)) 
	{
		return errCond(errMsgBfr, errMsgBfrLen, -1,
			"Write access denied for file %s", fileName);
	}

	if (stat(fileName, &statbuf) != 0) 
	{
		char bfr[256];
		bfr[0] = 0;
		strerror_r(errno, bfr, sizeof(bfr));
		return errCond(errMsgBfr, errMsgBfrLen, -1, "Unable to stat file %s. "
			"Error: %s", fileName, bfr);
	}

	if (!S_ISREG(statbuf.st_mode)) 
	{
		return errCond(errMsgBfr, errMsgBfrLen, -1, "%s is not a valid file "
			"type for modification", fileName);
	}

	cmpiutilStrNCpy(dirName, fileName, sizeof(dirName));
	if ((p = strrchr(dirName, '/'))) 
	{
		*p = 0;
	}
	else 
	{
		strcpy(dirName, ".");
	}

	origHdl = fopen(fileName, "r");
	if (!origHdl) 
	{
		return errCond(errMsgBfr, errMsgBfrLen, -1,
			"Failed to open file %s", fileName);
	}

	tempHdl = cmpiutilCreateTempFileInDir(dirName, tmpFileName, sizeof(tmpFileName));
	if (!tempHdl) 
	{
		fclose(origHdl);
		return errCond(errMsgBfr, errMsgBfrLen, -1, "Failed creating file");
	}

	if (cbs->fileOpened) 
	{
		modified = cbs->fileOpened(tempHdl);
	}

	for(p = cmpiutilReadLine(origHdl, &bufPtr, &bfrLen); p;
		p = cmpiutilReadLine(origHdl, &bufPtr, &bfrLen)) 
	{
		/* If line is blank, then just write it directly to the
		 * temp file
		 */
		if (!strlen(p)) 
		{
			fputs("\n", tempHdl);
			continue;
		}

		if (cbs->writeLine) 
		{
			cc = cbs->writeLine(p, tempHdl, errMsgBfr, errMsgBfrLen);
		}
		else 
		{
			if (fputs(p, tempHdl) == EOF) 
			{
				cmpiutilStrNCpy(errMsgBfr, "Failed writing to temp file",
					errMsgBfrLen);
				cc = -1;
			}
			else if (fputs("\n", tempHdl) == EOF) 
			{
				cmpiutilStrNCpy(errMsgBfr, "Failed writing to temp file",
					errMsgBfrLen);
				cc = -1;
			}
			else 
			{
				cc = 0;
			}
		}

		if (cc > 0) 
		{
			/* Line was ignored or modified */
			modified = 1;
		}
		else if (cc < 0) 
		{
			/* Callback gave us a condition we must abort on */
			aborted = 1;
			cmpiutilStrNCpy(errMsgBfr, "Modification aborted",
				errMsgBfrLen);
			break;
		}
	}

	if (cbs->fileClosing) 
	{
		if (cbs->fileClosing(tempHdl)) 
		{
			modified = 1;
		}
	}

	fclose(origHdl);
	fclose(tempHdl);
	if (bufPtr) 
	{
		/* Apparently cmpiutilReadLine allocation memory, so free it here. */
		free(bufPtr);
	}

	if (modified && !aborted) 
	{
		mode_t themode;
		char origFileName[PATH_MAX];

		/* Create the '.orig' file name */
		cmpiutilStrNCpy(origFileName, fileName, sizeof(origFileName));
		cmpiutilStrAppend(origFileName, ".orig", sizeof(origFileName));

		/* Remove previous backup file */
		remove(origFileName);

		/* Rename current file to back up */
		if (rename(fileName, origFileName) == -1) 
		{
			remove(tmpFileName);
			return -1;
		}

		/* Move modified file over to original file */
		if (rename(tmpFileName, fileName) == -1) 
		{
			/* Rename the .orig back to the real file name */
			rename(origFileName, fileName);
			remove(tmpFileName);
			return -1;
		}

		// Set the file permissions to what they were on the old file */
		themode = statbuf.st_mode &
			(S_ISUID | S_ISGID | S_ISVTX | S_IRWXU | S_IRWXG | S_IRWXO);
		chmod(fileName, themode);
	}
	else 
	{
		/* Nothing to do, so just delete the temp file */
		remove(tmpFileName);
	}

	if (cbs->operationComplete) 
	{
		cbs->operationComplete();
	}

	return (aborted) ? -1 : 0;
}

/****************************************************************************/
int cmpiutilModifyRollBack(
	const char* fileName,
	char* errMsgBfr,
	size_t errMsgBfrLen)
{
	char origFileName[PATH_MAX];
	struct stat statbuf;
	if(stat(fileName, &statbuf) != 0) 
	{
		return errCond(errMsgBfr, errMsgBfrLen, -1,
			"Failed to stat file %s", fileName);
	}
	origFileName[0] = '\0';
	cmpiutilStrNCpy(origFileName, fileName, sizeof(origFileName));
	cmpiutilStrAppend(origFileName, ".orig", sizeof(origFileName));
	if (access(origFileName, R_OK))
	{
		if (remove(fileName) == 0) 
		{
			rename(origFileName, fileName);

			// Set file permissions to what they were on the old file.
			mode_t themode = statbuf.st_mode &
				(S_ISUID | S_ISGID | S_ISVTX | S_IRWXU
					| S_IRWXG | S_IRWXO);
			chmod(fileName, themode);
		}
		else 
		{
			return errCond(errMsgBfr, errMsgBfrLen, -1,
				"Don't have write access to file %s",
					fileName);
		}
	}
	else 
	{
		return errCond(errMsgBfr, errMsgBfrLen, -1,
			"Don't have read access to file %s",
				origFileName);
	}
	return 0;
}

