/*******************************************************************************
* Copyright (C) 2005, 2006 Novell, Inc. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Novell, Inc. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Novell, Inc. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
* Author: Jon Carey (jcarey@novell.com)
******************************************************************************/
#include <cmpiutil/string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>

/*****************************************************************************/
uint32_t
cmpiutilStrHash(
		const unsigned char* str)
{
	uint32_t hash = 5381;
	int c;

	if (!str) 
	{
		return 0;
	}

	while ((c = *str++))
	{
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
	}

	return hash;
}

/*****************************************************************************/
const char*
cmpiutilStrEndsWith(
		const char* src,
		const char* arg)
{
	const char* p;
	size_t slen, alen;

	if (!src || !arg) 
	{
		errno = EINVAL;		/* Invalid parm */
		return NULL;
	}

	errno = 0;
	slen = strlen(src);
	alen = strlen(arg);

	if (alen > slen)
		return NULL;

	p = src + (slen - alen);
	return (strcmp(p, arg) == 0) ? p : NULL;
}

/*****************************************************************************/
const char*
cmpiutilStrStartsWith(
		const char* src,
		const char* arg)
{
	size_t slen, alen;

	if (!src || !arg) 
	{
		errno = EINVAL;		/* Invalid parm */
		return NULL;
	}

	errno = 0;
	slen = strlen(src);
	alen = strlen(arg);

	if (alen > slen)
		return NULL;

	return (strncmp(src, arg, alen) == 0) ? src : NULL;
}

/*****************************************************************************/
const char*
cmpiutilStrAppend(
		char* dest,
		const char* src,
		size_t maxResultLen)
{
	size_t dlen, slen, clen;

	if (!dest) 
	{
		errno = EINVAL;		/* Invalid parm */
		return NULL;
	}

	errno = 0;
	if (!src || !maxResultLen)
		return dest;

	dlen = strlen(dest);
	if (dlen >= maxResultLen)
		return dest;

	slen = strlen(src);

	clen = (maxResultLen - dlen) - 1;
	strncat(dest, src, clen);
	return dest;
}

/*****************************************************************************/
#define SPACEDELIMS " \f\n\r\t\v"
size_t
cmpiutilStrTokenCount(
		const char* src,
		const char* delims)
{
	const char* p = src;
	size_t tokcount = 0;

	if (!src) 
	{
		errno = EINVAL;
		return 0;
	}

	if (!delims || !*delims)
		delims = SPACEDELIMS;

	while(*p) 
	{
		// skip past delims
		while(strchr(delims, *p) && *p)
			p++;

		if (!*p)
			break;

		tokcount++;

		// Skip past non-delims
		while(!strchr(delims, *p) && *p)
			p++;
	}

	errno = 0;
	return tokcount;
}

/*****************************************************************************/
char**
cmpiutilStrTokenize(
		const char* src,
		const char* delims,
		size_t* numberOfElements)
{
	size_t tokcnt, psize;
	int i;
	char *saveptr, *rbfr, *str, **strs;

	if (!src || !*src) 
	{
		errno = EINVAL;
		return NULL;
	}

	if (!delims)
	{
		delims = SPACEDELIMS;
	}

	/* cmpiutilStrTokenCount sets/clears errno */
	tokcnt = cmpiutilStrTokenCount(src, delims);
	if (!tokcnt)
		return NULL;	/* No tokens. */

	psize = (tokcnt+1) * sizeof(char*);
	rbfr = (char*) malloc(psize + (sizeof(char) * (strlen(src)+1)));
	if (!rbfr) 
	{
		errno = ENOMEM;
		return NULL;	/* Mem allocation error */
	}

	if (numberOfElements) 
	{
		*numberOfElements = tokcnt;
	}

	memset(rbfr, 0, psize + (sizeof(char) * (strlen(src)+1)));
	strs = (char**) rbfr;	/* Pointer to string array (argv style) */
	str = rbfr + psize;		/* Pointer to string */
	strcpy(str, src);		/* copy string to bfr for tokenize */

	/* Start tokenizing string */
	char* p = strtok_r(str, delims, &saveptr);
	for(i = 0; p; i++) 
	{
		strs[i] = p;
		p = strtok_r(NULL, delims, &saveptr);
	}
	return strs;
}

/*****************************************************************************/
const char*
cmpiutilStrLTrim(
		char* src)
{
    if (src) 
	{
        char* p;
        for(p = src; *p && isspace(*p); p++)
		{
			/* Empty */
		}
        if (!*p)
		{
            *src = 0;
		}
        else if(p != src)
		{
            memmove(src, p, strlen(p)+1);
		}
    }
    return src;
}

/*****************************************************************************/
const char*
cmpiutilStrRTrim(
		char* src)
{
    if (src && *src) 
	{
        char *p = strchr(src, 0); /* find eos */
        for (--p; p >= src && isspace(*p); --p)
		{
            *p = 0;
		}
    }
    return src;
}

/*****************************************************************************/
const char*
cmpiutilStrTrim(
		char* src)
{
	cmpiutilStrRTrim(src);
	return cmpiutilStrLTrim(src);
}

/*****************************************************************************/
const char*
cmpiutilStrNCpy(
		char* dest,
		const char* src,
		size_t bytesToCopy)
{
	if (!dest || !bytesToCopy) 
	{
		errno = EINVAL;
		return NULL;
	}

	strncpy(dest, src, bytesToCopy-1);
	dest[bytesToCopy-1] = 0;
	return dest;
}

/*****************************************************************************/
cmpiutilStrArray*
cmpiutilStrArray_create()
{
	cmpiutilStrArray* pra = (cmpiutilStrArray*)calloc(1, sizeof(cmpiutilStrArray));
	if (!pra) {
		errno = ENOMEM;
		return NULL;
	}
	pra->strs = (char**) calloc(16, sizeof(char*));
	if (!pra->strs) {
		free(pra);
		errno = ENOMEM;
		return NULL;
	}
	pra->buf_size = 16;
	return pra;
}

/*****************************************************************************/
int
cmpiutilStrArray_destroy(cmpiutilStrArray* pra)
{
	size_t nstrs;

	if (!pra) {
		errno = EINVAL;
		return -1;
	}

	for(nstrs = 0; nstrs < pra->num_strings; nstrs++) {
		if (pra->strs[nstrs]) {
			free(pra->strs[nstrs]);
		}
	}
	free(pra->strs);
	free(pra);
	return 0;
}

/*****************************************************************************/
cmpiutilStrArray*
cmpiutilStrArray_append(cmpiutilStrArray* pra, const char* str)
{
	int slen = strlen(str);
	if (!pra) {
		errno = EINVAL;
		return NULL;
	}
	if (pra->num_strings == pra->buf_size-1) {
		size_t nsize = pra->buf_size + 16;
		char** nbuf = (char**) calloc(nsize, sizeof(char*));
		if (!nbuf) {
			errno = ENOMEM;
			return NULL;
		}
		memcpy(nbuf, pra->strs, sizeof(char*) * pra->num_strings);
		free(pra->strs);
		pra->strs = nbuf;
		pra->buf_size = nsize;
	}
	pra->strs[pra->num_strings] = (char*) malloc(slen+1);
	if (!pra->strs[pra->num_strings]) {
		errno = ENOMEM;
		return NULL;
	}
	strcpy(pra->strs[pra->num_strings], str);
	pra->num_strings++;
	return pra;
}

/*****************************************************************************/
int
cmpiutilStrArray_size(cmpiutilStrArray* pra)
{
	if (!pra) {
		errno = EINVAL;
		return -1;
	}
	return (int)pra->num_strings;
}

/*****************************************************************************/
int
cmpiutilStrArray_reset(cmpiutilStrArray* pra)
{
	size_t nstrs;

	if (!pra) {
		errno = EINVAL;
		return -1;
	}

	for(nstrs = 0; nstrs < pra->num_strings; nstrs++) {
		if (pra->strs[nstrs]) {
			free(pra->strs[nstrs]);
			pra->strs[nstrs] = NULL;
		}
	}
	pra->num_strings = 0;
	return 0;
}

/*****************************************************************************/
cmpiutilStrArray*
cmpiutilStrArray_set_element(cmpiutilStrArray* pra, const char* str, size_t ndx)
{
	int argslen = strlen(str);

	if (!pra) {
		errno = EINVAL;
		return NULL;
	}
	if (ndx >= pra->num_strings || !str) {
		errno = EINVAL;
		return NULL;
	}

	if (argslen > strlen(pra->strs[ndx])) {
		free(pra->strs[ndx]);
		pra->strs[ndx] = (char*) malloc(argslen+1);
		if (!pra->strs[ndx]) {
			errno = ENOMEM;
			return NULL;
		}
	}
	strcpy(pra->strs[ndx], str);
	return pra;
}

/*****************************************************************************/
const char*
cmpiutilStrArray_get_element(cmpiutilStrArray* pra, size_t ndx)
{
	if (!pra || ndx >= pra->num_strings) {
		errno = EINVAL;
		return NULL;
	}
	return pra->strs[ndx];
}

/*****************************************************************************/
cmpiutilStrBuf*
cmpiutilStrBuf_create()
{
	cmpiutilStrBuf* psbuf = (cmpiutilStrBuf*) malloc(sizeof(cmpiutilStrBuf));
	if (!psbuf) {
		errno = ENOMEM;
		return NULL;
	}
	psbuf->bfr = (char*) malloc(512);
	psbuf->bfr[0] = 0;
	psbuf->buf_len = 512;
	psbuf->str_len = 0;
	return psbuf;
}

/*****************************************************************************/
const char*
cmpiutilStrBuf_cat(cmpiutilStrBuf* psbuf, const char* str, size_t n)
{
	size_t slen;

	if (!psbuf) {
		errno = EINVAL;
		return NULL;
	}

	if (!str) {
		return psbuf->bfr;
	}

	slen = n ? n : strlen(str);
	if (!slen) {
		return psbuf->bfr;
	}
	if ((psbuf->str_len + slen) >= psbuf->buf_len) {
		int nsize = psbuf->str_len + slen + 512;
		char* nbuf = (char*) malloc(nsize);
		if (!nbuf) {
			errno = ENOMEM;
			return NULL;
		}
		strcpy(nbuf, psbuf->bfr);
		strncat(nbuf, str, slen);
		free(psbuf->bfr);
		psbuf->bfr = nbuf;
		psbuf->buf_len = nsize;
	}
	else {
		strncat(psbuf->bfr, str, slen);
	}
	psbuf->str_len = strlen(psbuf->bfr);
	return psbuf->bfr;
}

/*****************************************************************************/
int
cmpiutilStrBuf_reset(cmpiutilStrBuf* psbuf)
{
	if (!psbuf) {
		errno = EINVAL;
		return -1;
	}
	psbuf->bfr[0] = 0;
	psbuf->str_len = 0;
	return 0;
}

/*****************************************************************************/
const char*
cmpiutilStrBuf_set(cmpiutilStrBuf* psbuf, const char* str, size_t n)
{
	size_t slen;

	if (!psbuf) {
		errno = EINVAL;
		return NULL;
	}

	if (!str) {
		cmpiutilStrBuf_reset(psbuf);
		return psbuf->bfr;
	}

	slen = n ? n : strlen(str);
	psbuf->bfr[0] = 0;
	psbuf->str_len = slen;
	if (!slen) {
		return psbuf->bfr;
	}

	if (slen >= psbuf->buf_len) {
		int nsize = psbuf->str_len + 512;
		char* nbuf = (char*) malloc(nsize);
		if (!nbuf) {
			errno = ENOMEM;
			return NULL;
		}
		strncpy(nbuf, str, slen);
		nbuf[slen] = 0;
		free(psbuf->bfr);
		psbuf->bfr = nbuf;
		psbuf->buf_len = nsize;
	}
	else {
		strncpy(psbuf->bfr, str, slen);
		psbuf->bfr[slen] = 0;
	}
	return psbuf->bfr;
}

/*****************************************************************************/
int
cmpiutilStrBuf_destroy(cmpiutilStrBuf* psbuf)
{
	if (!psbuf) {
		errno = EINVAL;
		return -1;
	}
	if (psbuf->bfr) {
		free(psbuf->bfr);
	}
	free(psbuf);
	return 0;
}

